#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <gtk/gtk.h>
#include "icons/gmixer.xpm"
#include "icons/mix_logo.xpm"
#include "icons/mix_audio.xpm"
#include "icons/mix_bass.xpm"
#include "icons/mix_cd.xpm"
#include "icons/mix_ext.xpm"
#include "icons/mix_microphone.xpm"
#include "icons/mix_midi.xpm"
#include "icons/mix_recmon.xpm"
#include "icons/mix_treble.xpm"
#include "icons/mix_unknown.xpm"
#include "icons/mix_volume.xpm"

#define NAME_LEN 64

char *mixer_write_error = "Error writing mixer\n";
char *mixer_read_error = "Error reading mixer\n";

typedef struct volctrl
{
	gpointer *mixer_id;
	GtkWidget *vol_widget;
	GtkWidget *right_widget;
	GtkWidget *enable_widget;
	GtkWidget *rec_widget;
	GtkWidget *stereo_widget;
	GtkObject *vol_left;
	GtkObject *vol_right;
	char **xpm;
} VolumeControl;

/* rcfile = ~/.gmixerrc */
char rcfile[NAME_LEN];

/* mixer file descriptor */
int mixer_fd;

/* mixer masks */
int devmask;
int recmask;
int stereodevs;
GtkWidget *options;

char *vollabels[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_LABELS;

VolumeControl volctrls[SOUND_MIXER_NRDEVICES] =
{
	{(gpointer) "0",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_volume_xpm},
	{(gpointer) "1",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_bass_xpm},
	{(gpointer) "2",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_treble_xpm},
	{(gpointer) "3",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_midi_xpm},
	{(gpointer) "4",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_audio_xpm},
	{(gpointer) "5",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "6",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_ext_xpm},
	{(gpointer) "7",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_microphone_xpm},
	{(gpointer) "8",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_cd_xpm},
	{(gpointer) "9",  NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "10", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "11", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_recmon_xpm},
	{(gpointer) "12", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "13", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "14", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "15", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm},
	{(gpointer) "16", NULL, NULL, NULL, NULL, NULL, NULL, NULL, mix_unknown_xpm}
};

void read_config (void)
{
	int i, j;
	char buffer1[40], *buffer2;
	FILE *config_file;
	
	if ((config_file = fopen(rcfile, "r")) == NULL)
	{
		g_print ("Error reading config file\n");
		return;
	}
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		fgets (buffer1, 40, config_file);
		if ((buffer2 = (char *)strstr (buffer1, vollabels[i])) == NULL)
			return;
		if (buffer2[6] != '0')
		{
			if ((buffer2 = (char *)strstr (buffer1, "enable")) != NULL)
				if (buffer2[7] == '0' &&  ((devmask >> i) & 0x1) == 0x1)
					gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(volctrls[i].enable_widget), FALSE);
			if ((buffer2 = (char *)strstr (buffer1, "stereo")) != NULL)
				if (buffer2[7] == '0' && ((stereodevs >> i) & 0x1) == 0x1)
					gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(volctrls[i].stereo_widget), FALSE);
			if ((buffer2 = (char *)strstr (buffer1, "rec")) != NULL)
				if (buffer2[4] == '0' && ((recmask >> i) & 0x1) == 0x1)
					gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(volctrls[i].rec_widget), FALSE);
		}
	}
	fclose (config_file);
}

void write_config (void)
{
	int i;
	FILE *config_file;
	
	if ((config_file = fopen(rcfile, "w")) == NULL)
	{
		g_print ("Error writing config file\n");
		return;
	}
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		fprintf (config_file, "%s", vollabels[i]);
		if (((devmask >> i) & 0x1) == 0x1)
		{
			fprintf (config_file, " 1");
			if (GTK_TOGGLE_BUTTON (volctrls[i].enable_widget)->active)
				fprintf (config_file, " enable 1");
			else
				fprintf (config_file, " enable 0");
			if (((stereodevs >> i) & 0x1) == 0x1)
				if (GTK_TOGGLE_BUTTON (volctrls[i].stereo_widget)->active)
					fprintf (config_file, " stereo 1");
				else
					fprintf (config_file, " stereo 0");
			if (((recmask >> i) & 0x1) == 0x1)
				if (GTK_TOGGLE_BUTTON (volctrls[i].rec_widget)->active)
					fprintf (config_file, " rec 1");
				else
					fprintf (config_file, " rec 0");
			fprintf (config_file, "\n");
		}
		else 
			fprintf (config_file, " 0\n");
	}
	fclose (config_file);
}

void delete_event (GtkWidget *widget, gpointer *data)
{
	write_config ();
	gtk_main_quit ();
}

void options_show (GtkWidget *widget, gpointer *data)
{
	gtk_widget_show (options);
}

void options_hide (GtkWidget *widget, gpointer *data)
{
	gtk_widget_hide (options);
}

void option_enable (GtkWidget *widget, gpointer *data)
{
	int i;
	i = atoi((char *)data);
	if (GTK_TOGGLE_BUTTON (widget)->active)
		gtk_widget_show (volctrls[i].vol_widget);
	else
	{
		gtk_adjustment_set_value (GTK_ADJUSTMENT(volctrls[i].vol_left), 100);
		if (((stereodevs >> i) & 0x1) == 0x1)
			gtk_adjustment_set_value (GTK_ADJUSTMENT(volctrls[i].vol_right), 100);
		gtk_widget_hide (volctrls[i].vol_widget);
	}
}

void option_rec (GtkWidget *widget, gpointer *data)
{
	int i, rec, recsrc;
	i = atoi((char *)data);
	if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &recsrc) == -1)
		g_print (mixer_read_error);
	if (GTK_TOGGLE_BUTTON (widget)->active)
		recsrc |=  (0x1 << i);
	else
		recsrc &= 0xffff - (0x1 << i);
	if (ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &recsrc) == -1)
		g_print (mixer_write_error);
	if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &recsrc) == -1)
		g_print (mixer_read_error);
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
		if (((devmask >> i) & 0x1) == 0x1 && ((recmask >> i) & 0x1) == 0x1)
			gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(volctrls[i].rec_widget), (recsrc >> i) & 0x1);
}

void option_stereo (GtkWidget *widget, gpointer *data)
{
	int i;
	i = atoi((char *)data);
	gtk_adjustment_set_value (GTK_ADJUSTMENT(volctrls[i].vol_right), GTK_ADJUSTMENT(volctrls[i].vol_left)->value);
	if (GTK_TOGGLE_BUTTON (widget)->active)
	{
		gtk_widget_show (volctrls[i].right_widget);
	}
	else
		gtk_widget_hide (volctrls[i].right_widget);
}

void volume_changed (GtkWidget *widget, gpointer *data)
{
	int i, v_left, v_right, v;
	i = atoi((char *)data);
	v_left = 100 - GTK_ADJUSTMENT(volctrls[i].vol_left)->value;
	if (((stereodevs >> i) & 0x1) == 0x1 && GTK_TOGGLE_BUTTON(volctrls[i].stereo_widget)->active)
		v_right = 100 - GTK_ADJUSTMENT(volctrls[i].vol_right)->value;
	else
		v_right = v_left;
	v = (v_right << 8) + v_left;
	if (ioctl(mixer_fd, MIXER_WRITE(i), &v) == -1)
		g_print (mixer_write_error);
}

void update_gui(void)
{
	int i, recsrc, volume;
	if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &recsrc) == -1)
		g_print (mixer_read_error);
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if (((devmask >> i) & 0x1) == 0x1)
		{
			if (((recmask >> i) & 0x1) == 0x1)
				gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(volctrls[i].rec_widget), (recsrc >> i) & 0x1);
			if (ioctl(mixer_fd, MIXER_READ(i), &volume) == -1)
				g_print (mixer_read_error);
			gtk_adjustment_set_value (GTK_ADJUSTMENT(volctrls[i].vol_left), 100 - volume & 0xff);
			if (((stereodevs >> i) & 0x1) == 0x1)
				gtk_adjustment_set_value (GTK_ADJUSTMENT(volctrls[i].vol_right), 100 - (volume >> 8) & 0xff);
		}
	}	
}

void focus_in_event (GtkWidget *widget, gpointer *data)
{
	update_gui ();
}

void create_gui ()
{
	int i;
	GtkWidget *window, *box1, *box2, *box3, *control;
	GtkWidget *option_box, *frame, *picture;
	GdkPixmap *pixmap;
	GdkBitmap *bitmap;
	GtkStyle *style;
	GtkTooltips *tooltips;
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	options = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_policy (GTK_WINDOW(window), FALSE, FALSE, TRUE);
	gtk_window_set_policy (GTK_WINDOW(options), FALSE, FALSE, TRUE);
	gtk_window_set_title (GTK_WINDOW (window), "gmixer");
	gtk_window_set_title (GTK_WINDOW (options), "Options");
	gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		GTK_SIGNAL_FUNC (delete_event), NULL);
	gtk_signal_connect (GTK_OBJECT (window), "focus_in_event",
		GTK_SIGNAL_FUNC (focus_in_event), NULL);
	gtk_signal_connect (GTK_OBJECT (options), "delete_event",
		GTK_SIGNAL_FUNC (options_hide), NULL);
	gtk_widget_realize (window);
	gtk_container_border_width (GTK_CONTAINER (window), 5);
	gtk_container_border_width (GTK_CONTAINER (options), 5);
	box1 = gtk_hbox_new (FALSE, 1);
	option_box = gtk_vbox_new (FALSE, 0);
	style = gtk_widget_get_style (window);
	gtk_container_add (GTK_CONTAINER (window), box1);
	gtk_container_add (GTK_CONTAINER (options), option_box);
	tooltips = gtk_tooltips_new();
	box2 = gtk_vbox_new (TRUE, 0);
	gtk_box_pack_start (GTK_BOX (box1), box2, TRUE , TRUE, 0);
	pixmap = gdk_pixmap_create_from_xpm_d (window->window, &bitmap,
		&style->bg[GTK_STATE_NORMAL], mix_logo_xpm);
	picture = gtk_pixmap_new (pixmap, bitmap);
	control = gtk_button_new ();
	gtk_container_add (GTK_CONTAINER (control), picture);
	gtk_box_pack_start (GTK_BOX (box2), control, FALSE , FALSE, 0);
	gtk_signal_connect (GTK_OBJECT (control), "clicked",
		GTK_SIGNAL_FUNC (options_show), NULL);
	gtk_tooltips_set_tip (tooltips, control, "Options", "");
	gtk_widget_show (picture);
	gtk_widget_show (control);
	gtk_widget_show (box2);
	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
		if (((devmask >> i) & 0x1) == 0x1)
		{
			volctrls[i].vol_widget = gtk_vbox_new (FALSE, 0);
			pixmap = gdk_pixmap_create_from_xpm_d (window->window, &bitmap,
				&style->bg[GTK_STATE_NORMAL], volctrls[i].xpm);
			picture = gtk_pixmap_new (pixmap, bitmap);
			gtk_box_pack_start (GTK_BOX (volctrls[i].vol_widget), picture, TRUE, FALSE, 0);
			gtk_widget_show (picture);
			box3 = gtk_hbox_new (FALSE, 0);
			gtk_box_pack_start (GTK_BOX (volctrls[i].vol_widget), box3, TRUE, FALSE, 0);
			volctrls[i].vol_left = gtk_adjustment_new (100.0, 0.0, 100.0, 1.0, 5.0, 1.0);
			control = gtk_vscale_new (GTK_ADJUSTMENT(volctrls[i].vol_left));
			gtk_range_set_update_policy( GTK_RANGE(control), GTK_UPDATE_CONTINUOUS );
		       	gtk_scale_set_draw_value( GTK_SCALE(control), FALSE );
			gtk_signal_connect (GTK_OBJECT (volctrls[i].vol_left), "value_changed",
		        	GTK_SIGNAL_FUNC (volume_changed), volctrls[i].mixer_id);
			gtk_box_pack_start (GTK_BOX (box3), control, TRUE, FALSE, 0);
			gtk_tooltips_set_tip (tooltips, control, vollabels[i], "");
			gtk_widget_show (control);
			gtk_box_pack_start (GTK_BOX (box1), volctrls[i].vol_widget, TRUE, TRUE, 0);
			frame = gtk_frame_new (vollabels[i]);
			gtk_box_pack_start (GTK_BOX (option_box), frame, TRUE, TRUE, 0);
			box2 = gtk_hbox_new (FALSE, 0);
			gtk_container_add (GTK_CONTAINER (frame), box2);
			volctrls[i].enable_widget = gtk_check_button_new_with_label ("Enable");
			gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (volctrls[i].enable_widget), TRUE);
			gtk_signal_connect (GTK_OBJECT (volctrls[i].enable_widget), "clicked",
				GTK_SIGNAL_FUNC (option_enable), volctrls[i].mixer_id);
			gtk_box_pack_start (GTK_BOX (box2), volctrls[i].enable_widget, FALSE, FALSE, 0);
			gtk_widget_show (volctrls[i].enable_widget);
			if (((stereodevs >> i) & 0x1) == 0x1) {
				volctrls[i].vol_right = gtk_adjustment_new (100.0, 0.0, 100.0, 1.0, 5.0, 1.0);
				volctrls[i].right_widget = gtk_vscale_new (GTK_ADJUSTMENT(volctrls[i].vol_right));
				gtk_range_set_update_policy( GTK_RANGE(volctrls[i].right_widget), GTK_UPDATE_CONTINUOUS);
			       	gtk_scale_set_draw_value( GTK_SCALE(volctrls[i].right_widget), FALSE );
				gtk_signal_connect (GTK_OBJECT (volctrls[i].vol_right), "value_changed",
					GTK_SIGNAL_FUNC (volume_changed), volctrls[i].mixer_id);
				gtk_box_pack_start (GTK_BOX (box3), volctrls[i].right_widget, TRUE, FALSE, 0);
				gtk_tooltips_set_tip (tooltips, volctrls[i].right_widget, vollabels[i], "");
				gtk_widget_show (volctrls[i].right_widget);
				volctrls[i].stereo_widget = gtk_check_button_new_with_label ("Stereo");
				gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (volctrls[i].stereo_widget), TRUE);
				gtk_box_pack_start (GTK_BOX (box2), volctrls[i].stereo_widget, FALSE, FALSE, 0);
				gtk_signal_connect (GTK_OBJECT (volctrls[i].stereo_widget), "clicked",
					GTK_SIGNAL_FUNC (option_stereo), volctrls[i].mixer_id);
				gtk_widget_show (volctrls[i].stereo_widget);
			}
			if (((recmask >> i) & 0x1) == 0x1) {
				volctrls[i].rec_widget = gtk_check_button_new_with_label ("Rec");
				gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (volctrls[i].rec_widget), TRUE);
				gtk_box_pack_start (GTK_BOX (box2), volctrls[i].rec_widget, FALSE, FALSE, 0);
				gtk_signal_connect (GTK_OBJECT (volctrls[i].rec_widget), "clicked",
					GTK_SIGNAL_FUNC (option_rec), volctrls[i].mixer_id);
				gtk_widget_show (volctrls[i].rec_widget);
			}
			gtk_widget_show (box3);
			gtk_widget_show (volctrls[i].vol_widget);
			gtk_widget_show (box2);
			gtk_widget_show (frame);
		}
	control = gtk_button_new_with_label ("Close");
	gtk_box_pack_start (GTK_BOX (option_box), control, TRUE, TRUE, 0);
	gtk_signal_connect (GTK_OBJECT (control), "clicked",
		GTK_SIGNAL_FUNC (options_hide), NULL);
	gtk_widget_show (control);
	gtk_widget_show (box1);
	gtk_widget_show (option_box);
	pixmap = gdk_pixmap_create_from_xpm_d (window->window, &bitmap,
		&style->bg[GTK_STATE_NORMAL], gmixer_xpm);
	gdk_window_set_icon (window->window, NULL, pixmap, bitmap);
	gtk_widget_show (window);
}

void read_mixer_caps (void)
{
	if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1)
		devmask = 0xffff;	/* Assume all are supported */
	if (ioctl(mixer_fd, SOUND_MIXER_READ_RECMASK, &recmask) == -1)
		recmask = 0xffff;
	if (ioctl(mixer_fd, SOUND_MIXER_READ_STEREODEVS, &stereodevs) == -1)
		stereodevs = 0xffff;
}

void usage(char *prog_name)
{
	g_print ("Usage:\n    %s [options]\nOptions:\n    -h, --help\n\tPrints a usage message and exits.\n    -d <device>, --device <device>\n\tUse the given device. Default is /dev/mixer.\n    -o <file>, --options-file <file>\n\tUse the given options file. Default is $HOME/.gmixerrc.\n", prog_name);
	exit(0);
}

int main (int argc, char *argv[])
{
	char *homedir;
	char *mixer_dev;
	char *default_dev = "/dev/mixer";
	int i;
	g_print ("gmixer 0.98 (22.2.1999)\nCopyleft (C) 1998,1999 - Sergey Kiselev <sergey@junior.technion.ac.il>\n");
	homedir = getenv("HOME");
	snprintf (rcfile, NAME_LEN, "%s/.gmixerrc", homedir);
	mixer_dev = default_dev;
	if (argc > 1)
		for (i = 1; i < argc; i++)
		{
			if (!strcmp ("-h", argv[i]) ||
			    !strcmp ("--help", argv[i]))
				usage (argv[0]);
			else if (i + 1 < argc && (! strcmp("-d", argv[i]) ||
			    !strcmp ("--device", argv[i])))
				mixer_dev = argv[++i];
			else if (i + 1 < argc && (! strcmp("-o", argv[i]) ||
			    !strcmp ("--options-file", argv[i])))
				snprintf (rcfile, NAME_LEN, "%s", argv[++i]);
			else
				usage (argv[0]);
		}
	mixer_fd = open (mixer_dev, O_RDWR, 0);
	if (mixer_fd < 0)
	{
 		g_print ("Error opening mixer device %s\n", mixer_dev);
		exit (1);
	}
	read_mixer_caps ();
	gtk_init (&argc, &argv);
	create_gui ();
	update_gui ();	
	read_config ();
	gtk_main ();
	close (mixer_fd);
	return 0;
}
